unit searches;

interface

{/*
 *	SEARCH.C
 *	Tom Kerrigan's Simple Chess Program (TSCP)
 *
 *	Copyright 1997 Tom Kerrigan
 */}
// Pascal version


uses Windows,sysUtils,board,evals;


procedure think(output:integer);
function move_str(m:TMove):pchar;
function reps:integer;

var
{/* a "triangular" PV array */}
  pv:array[0..Max_Ply-1,0..Max_Ply-1] of TMove;
  pv_length:array[0..Max_Ply-1] of integer;
  follow_pv:boolean;

var
{/* the time when the engine starts searching, and when it should stop */}
  start_time:longint;
  stop_time:longint;

{/* the engine will search for max_time milliseconds or until it finishes
   searching max_depth ply. */}
   max_time:longint;
   max_depth:integer;

  nodes:longint;  {/* the number of nodes we've searched */}

var
  stop_Search:boolean=False;


implementation

{/* checkup() is called once in a while during the search. */}

procedure checkup;
begin
	{/* is the engine's time up? if so, longjmp back to the
	   beginning of think() */}
	if (GetTickCount >= Dword(stop_time)) then
		stop_search := TRUE;
end;

{/* sort_pv() is called when the search function is following
   the PV (Principal Variation). It looks through the current
   ply's move list to see if the PV move is there. If so,
   it adds 10,000,000 to the move's score so it's played first
   by the search function. If not, follow_pv remains FALSE and
   search() stops calling sort_pv(). */}

procedure sort_pv;
var
	i:integer;
begin
	follow_pv := FALSE;
	for i := gen_begin[ply] to gen_end[ply]-1 do
		if longint(gen_dat[i].m) = longint(pv[0][ply]) then
    begin
			follow_pv := TRUE;
			inc(gen_dat[i].score,10000000);
			exit;
		end
end;

{/* sort() searches the current ply's move list from 'from'
   to the end to find the move with the highest score. Then it
   swaps that move and the 'from' move so the move with the
   highest score gets searched next, and hopefully produces
   a cutoff. */}

procedure sort(from:integer);
var
	i:integer;
	bs:longint;  {/* best score */}
	bi:integer;  {/* best i */}
	g:TGen;
begin
	bs := -1;
	bi := from;
	for i := from to gen_end[ply]-1 do
		if (gen_dat[i].score > bs) then
    begin
			bs := gen_dat[i].score;
			bi := i;
		end;
	g := gen_dat[from];
	gen_dat[from] := gen_dat[bi];
	gen_dat[bi] := g;
end;

{/* reps() returns the number of times that the current
   position has been repeated. Thanks to John Stanback
   for this clever algorithm. */}

function reps:integer;
var
	i:integer;
	b:array[0..63] of integer;
	c:integer;  {/* count of squares that are different from
				   the current position */}
	r:integer;  {/* number of repetitions */}

begin
	c := 0;
	r := 0;

	{/* is a repetition impossible? */}
	if (fifty <= 3) then
  begin
    result:=0;
    exit;
  end;
	for i := 0 to 63 do
		b[i] := 0;

	{/* loop through the reversible moves */}
	for i := hply - 1 downto  hply - fifty - 1 do with hist_dat[i] do
  begin
		inc(b[m.from]);
		if (b[m.from] = 0) then
			dec(c)
		else
			inc(c);
		dec(b[m.toNdx]);
		if (b[m.toNdx] = 0) then
			dec(c)
		else
			inc(c);
		if (c = 0) then
			inc(r)
	end;

	result:=r;
end;


{/* quiesce() is a recursive minimax search function with
   alpha-beta cutoffs. In other words, negamax. It basically
   only searches capture sequences and allows the evaluation
   function to cut the search off (and set alpha). The idea
   is to find a position where there isn't a lot going on
   so the static evaluation function will work. */}

function quiesce(alpha,beta:integer):integer;
var
	i, j, x:integer;
begin
	inc(nodes);

	{/* do some housekeeping every 1024 nodes */}
	if ((nodes and 1023) = 0) then
  begin
		checkup();
    if (stop_search=True) then
    begin
      Result:=0;
      exit;
    end;
  end;

	pv_length[ply] := ply;

	{/* are we too deep? */}
	if (ply >= MAX_PLY - 1) then
  begin
    result:=eval;
    exit;
  end;
	if (hply >= HIST_STACK - 1) then
  begin
    result:=eval;
    exit;
  end;

	{/* check with the evaluation function */}
	x := eval;
	if (x >= beta) then
  begin
    result:=beta;
    exit;
  end;

	if (x > alpha) then
		alpha := x;

	gen_caps;
	if (follow_pv=True) then  {/* are we following the PV? */}
		sort_pv;

	{/* loop through the moves */}
  for i:=gen_begin[ply] to gen_end[ply]-1 do
  begin
		sort(i);
		if not makemove(gen_dat[i].m) then
			continue;
		x := -quiesce(-beta, -alpha);
		takeback;
		if (x > alpha) then
    begin
			if (x >= beta) then
      begin
        result:=beta;
        exit;
      end;

			alpha := x;

			{/* update the PV */}
			pv[ply][ply] := gen_dat[i].m;
			for j := ply + 1 to pv_length[ply + 1]-1 do
				pv[ply][j] := pv[ply + 1][j];
			pv_length[ply] := pv_length[ply + 1];
		end
	end;
  result:=alpha;
end;


{/* search() does just that, in negamax fashion */}

function search(alpha, beta, depth:integer):integer;
var
	i, j, x:integer;
	c, f:Bool;

begin
	{/* we're as deep as we want to be; call quiesce() to get
	   a reasonable score and return it. */}
	if (depth=0) then
  begin
		result:=quiesce(alpha,beta);
    exit;
  end;
	inc(nodes);

	{/* do some housekeeping every 1024 nodes */}
	if ((nodes and 1023) = 0) then
  begin
		checkup();
    if (stop_search=True) then
    begin
      Result:=0;
      exit;
    end;
  end;

	pv_length[ply] := ply;

	{/* if this isn't the root of the search tree (where we have
	   to pick a move and can't simply return 0) then check to
	   see if the position is a repeat. if so, we can assume that
	   this line is a draw and return 0. */}
	if ((ply<>0) and (reps<>0)) then
  begin
    result:=0;
    exit;
  end;
	{/* are we too deep? */}
	if (ply >= MAX_PLY - 1) then
  begin
    result:=eval();
    exit;
  end;
	if (hply >= HIST_STACK - 1) then
  begin
    result:=eval();
    exit;
  end;

	{/* are we in check? if so, we want to search deeper */}
	c := in_check(side);
	if (c=TRUE) then
		inc(depth);
	gen;
	if (follow_pv=TRUE) then  {/* are we following the PV? */}
		sort_pv;
	f := FALSE;

	{/* loop through the moves */}
  for i:=gen_begin[ply] to gen_end[ply]-1 do
  begin
		sort(i);
		if not makemove(gen_dat[i].m) then
			continue;
		f := TRUE;
		x := -search(-beta, -alpha, depth - 1);
		takeback;
		if (x > alpha) then
    begin

			{/* this move caused a cutoff, so increase the history
			   value so it gets ordered high next time we can
			   search it */}
			inc(history[gen_dat[i].m.from][gen_dat[i].m.toNdx],depth);
			if (x >= beta) then
      begin
        result:=beta;
        exit;
      end;
			alpha := x;

			{/* update the PV */}
			pv[ply][ply] := gen_dat[i].m;
      for j:=ply+1 to pv_length[ply+1]-1 do
				pv[ply][j] := pv[ply + 1][j];
			pv_length[ply] := pv_length[ply + 1];
		end
	end;

	{/* no legal moves? then we're in checkmate or stalemate */}
	if (f=False) then
   begin
		if (c=True) then
    begin
      result:= -10000 + ply;
      exit;
    end
		else
    begin
      result:=0;
      exit;
    end
	end;

	{/* fifty move draw rule */}
	if (fifty >= 100) then
  begin
    result:=0;
    exit;
  end;
  result:=alpha;
end;

{/* move_str returns a string with move m in coordinate notation */}

function move_str(m:TMove):pchar;
const
  str:string='';
var
	c:char;
begin
	if ppmPromote in m.bits then
  begin
		case (m.promote) of
			KNIGHT:
				c := 'n';
			BISHOP:
				c := 'b';
			ROOK:
				c := 'r';
			else
				c := 'q';
		end;
		Str:=Format('%s%d%s%d%s',[
				char((m.from and 7) + ord('a')),
				8 - (m.from shr 3),
				char((m.toNdx and 7) + ord('a')),
				8 - (m.toNdx shr 3),
				c]);
	end
	else
		Str:=Format('%s%d%s%d',[
				char((m.From and 7) + ord('a')),
				8 - (m.from shr 3),
				char((m.toNdx and 7) + ord('a')),
				8 - (m.toNdx shr 3)]);
	result:=Pchar(str);
end;


{* think() calls search() iteratively. Search statistics
   are printed depending on the value of output:
   0 := no output
   1 := normal output
   2 := xboard format output */}

procedure think(output:integer);
var
	i, j, x:integer;
begin
	{/* some code that lets us longjmp back here and return
	   from think() when our time is up */}

	start_time := GetTickCount;
	stop_time := start_time + max_time;

	ply := 0;
	nodes := 0;
  fillchar(pv,sizeof(pv),0);
  fillchar(history,sizeof(history),0);
	if (output = 1) then
		writeln('ply      nodes  score  pv');
  for i:=1 to max_depth do
  begin
		follow_pv := TRUE;
		x := search(-10000, 10000, i);
    if (stop_search=True) then
    begin
      {/* make sure to take back the line we were searching */}
      while (ply<>0) do
        takeback;
      exit;
    end;
		if (output = 1) then
			write(i:3,' ', nodes:9,' ', x:5)
		else if (output = 2) then
			write(i,' ', x,' ', (GetTickCount - Dword(start_time)) div 10,' ', nodes);
		if (output=1) then
    begin
      for j:=0 to pv_length[0]-1 do
				write(' ', move_str(pv[0][j]));
			writeln;
		end
	end
end;





end.
